/**
 * The Render Engine
 * KeyboardInputComponent
 *
 * @fileoverview An extension of the input component to handle touch inputs from
 *               devices which support them.
 *
 * @author: Brett Fattori (brettf@renderengine.com)
 * @author: $Author: bfattori $
 * @version: $Revision: 1555 $
 *
 * Copyright (c) 2011 Brett Fattori (brettf@renderengine.com)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */

// The class this file defines and its required classes
R.Engine.define({
    "class":"R.struct.Touch",
    "requires":[
        "R.math.Point2D"
    ]
});

/**
 * @class A data object which represents a touch event generated by {@link R.components.TouchInput}.
 *        Since touches are handled differently across platforms, this class aims to
 *        normalize what is returned and allow a game to handle the input without concern
 *        for platform.
 *
 * @param touch {Event} The touch from the event object
 * @constructor
 * @description A touch data object.
 */
R.struct.Touch = Base.extend(/** @scope R.struct.Touch.prototype */{

    touchX:null,
    touchY:null,

    /**
     * @private
     */
    constructor:function (touch) {
        this.touchX = touch.pageX;
        this.touchY = touch.pageY;
    },

    /**
     * Get the X coordinate of the touch
     * @return {Number} The X coordinate of the touch
     */
    getX:function () {
        return this.touchX;
    },

    /**
     * Get the Y coordinate of the touch
     * @return {Number} The Y coordinate of the touch
     */
    getY:function () {
        return this.touchY;
    },

    /**
     * Get a {@link R.math.Point2D} which represents the location of the touch
     * @return {R.math.Point2D} The coordinates of the touch
     */
    get:function () {
        return R.math.Point2D.create(this.touchX, this.touchY);
    }

}, /** @scope R.struct.Touch.prototype */{
    /**
     * Get the class name of this object
     *
     * @return {String} "R.struct.Touch"
     */
    getClassName:function () {
        return "R.struct.Touch";
    }
});
